;ATTiny2313 Bootloader
;
;Features:
; - automatic switching to main program if nothing is received within ~2s
; - transparent to the application (except the smaller available flash size)
; - located at the end of the flash memory
; - automatic modification of the application while uploading
; - AVR910 compatible protocol (see below for supported commands)
;
;Requirements:
; - ATTiny2313 running at any frequency which can provide the selected BAUD rate
;     in double speed mode
; - RS232-Interface connected to PD0/RxD and PD1/TxD in the target application
; - The main (user) program MUST have a rjmp command at 0x000!
;
;Version:
;  v0.1, 2005-06-26 (Arne Rossius)
;  v0.2, 2005-06-29 (Arne Rossius)
;    * first working version
;  v0.3, 2005-06-30 (Arne Rossius)
;    * moved to the end of the flash, does not work yet!
;  v0.7, 2006-01-29 (Arne Rossius)
;    * loads of bugs fixed
;  v0.8  2007-04-23 (Arne Rossius)
;    * working version at the end of the flash
;
;Programming commands (subset of avr910.asm from Atmel):
;*                                     +-------------+------------+------+
;*  Commands                           | Host writes | Host reads |      |
;*  --------                           +-----+-------+------+-----+      |
;*                                     | ID  | data  | data |     | Note |
;* +-----------------------------------+-----+-------+------+-----+------+
;* | Enter programming mode            | 'P' |       |      | 13d |   1  |
;* | Report autoincrement address      | 'a' |       |      | 'Y' |      |
;* | Set address                       | 'A' | ah al |      | 13d |   2  |
;* | Write program memory, low byte    | 'c' |    dd |      | 13d |   3  |
;* | Write program memory, high byte   | 'C' |    dd |      | 13d |   3  |
;* | Issue Page Write                  | 'm' |       |      | 13d |      |
;* | Read program memory               | 'R' |       |dd(dd)|     |   4  |
;* | Write data memory                 | 'D' |    dd |      | 13d |      |
;* | Read data memory                  | 'd' |       |   dd |     |      |
;* | Chip erase (excluding bootloader) | 'e' |       |      | 13d |      |
;* | Leave programming mode            | 'L' |       |      | 13d |   5  |
;* | Select device type (5E only)      | 'T' |    dd |      | 13d |   6  |
;* | Return signature bytes (1E910A)   | 's' |       | 3*dd |     |      |
;* | Return supported device code (5E) | 't' |       | n*dd | 00d |   7  |
;* | Return software ident. "UTSRQPO"  | 'S' |       | s[7] |     |   8  |
;* | Return sofware version            | 'V' |       |dd dd |     |   9  |
;* | Return hardware version (0.0)     | 'v' |       |dd dd |     |   9  |
;* | Return programmer type ("S")      | 'p' |       |   dd |     |  10  |
;* +-----------------------------------+-----+-------+------+-----+------+

.equ	VerMaj = 0 ;major version number
.equ	VerMin = 8 ;minor version number
.equ	Freq = 8000000 ;clock frequency
.equ	Baud = 19200 ;rs232 baud rate
.equ	DevCode = 0x5E ;AVR910 device code for Tiny2313

.equ	EEPROM = 0 ;enable/disable EEPROM programming (disable to save space)

;do not change this unless you have changed the bootloader code!
.if ( EEPROM )
	.equ	BLSIZE = 15 ;boot loader size in pages (1 page = 16 words) with EEPROM code
.else
	.equ	BLSIZE = 14 ;boot loader size in pages (1 page = 16 words) without EEPROM code
.endif

.equ	Count1_init = Freq / 500000 ;experimentally determined value

;data to write to the page buffer (must be R0/R1)
.def	DataL = R0
.def	DataH = R1

.def	temp = R16
.def	temp2 = R17

;instruction for 0x000
.def	InstrL = R18
.def	InstrH = R19

;countdown
.def	count = R22
.def	Count1 = R23
.def	Count2 = R24
.def	Count3 = R25

;address backup
.def	Addr2L = R28
.def	Addr2H = R29

;flash/EEPROM address (must be the Z register)
.def	AddrL = R30
.def	AddrH = R31

.include "../include/tn2313def.inc"
.cseg
.org 0x000
	rjmp	bl_start
.org 0x001

progstart:
	
	; ***************************************************
	; ******* MAIN PROGRAM WILL BE INSERTED HERE! *******
	; ***************************************************
	
.org (FLASHEND+1)-(BLSIZE*PAGESIZE)
bl_start:
	;disable interrupts
	cli
	
	;stack pointer setup
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;enable pullup for RxD
	ldi	temp, 0x01
	out	PORTD, temp
	
	;configure UART
	ldi	temp, 1<<U2X ;double speed mode
	out	UCSRA, temp
	ldi	temp, 1<<RXEN | 1<<TXEN ;RX on, TX on
	out	UCSRB, temp
	ldi	temp, Freq / ((8 * Baud) - 1)
	out	UBRRL, temp
	
	;set initial address
	ldi	AddrL, 0x00
	ldi	AddrH, 0x00
	
	;EEPROM: atomic write (erase + write)
	out	EECR, AddrL
	
	;reset & enable timeout (jump to main program) countdown
	rcall	countdown_reset
	ldi	count, 1
	
bl_loop:
	rcall	getc
	
;****************************** controller ******************************
	
	cpi	temp, 's'		;'s' read signature bytes (always 1E910A)
	brne	bl_s
	ldi	temp, 0x0A
	rcall	putc
	ldi	temp, 0x91
	rcall	putc
	ldi	temp, 0x1E
	rjmp	ret_put
bl_s:

;****************************** address ******************************
	
	cpi	temp, 'a'		;'a' report autoincrement address
	brne	bl_a
	ldi	temp, 'Y'
	rjmp	ret_put

bl_a:
	cpi	temp, 'A'		;'A' set address
	brne	bl_AA
	rcall	getc ;MSB first
	mov	AddrH, temp
	rcall	getc
	mov	AddrL, temp
	lsl	AddrL ;convert word address to byte address
	rol	AddrH
	rjmp	ret_ok
bl_AA:

;****************************** EEPROM memory ******************************
	
.if ( EEPROM )
	cpi	temp, 'd'		;'d' read data memory
	brne	bl_d
	mov	temp, AddrL
	lsr	temp ;eeprom uses bytes, not words
bl_d1:
	out	EEAR, temp
	sbi	EECR, EERE
	in	temp, EEDR
	adiw	AddrH:AddrL, 2 ;increment Addr
	rjmp	ret_put
bl_d:
	
	cpi	temp, 'D'		;'D' write data memory
	brne	bl_DD
	mov	temp, AddrL
	lsr	temp ;eeprom uses bytes, not words
	out	EEAR, temp
	rcall	getc
	out	EEDR, temp
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
bl_DD1:
	sbic	EECR, EEPE ;wait until EEPROM is ready
	rjmp	bl_DD1
	adiw	AddrH:AddrL, 2 ;increment Addr
	rjmp	ret_ok
bl_DD:
.endif

;****************************** flash memory ******************************
	
	cpi	temp, 'R'		;'R' read program memory
	brne	bl_RR
	movw	Addr2H:Addr2L, AddrH:AddrL ;backup address
	;set Addr to addr0 and convert rjmp if Addr is 0x000
	cpi	AddrL, 0x00
	cpc	AddrH, AddrL
	brne	bl_RR1
	ldi	AddrL, LOW(addr0*2)
	ldi	AddrH, HIGH(addr0*2)
	lpm	InstrL, Z+
	lpm	temp, Z+
	andi	temp, 0x0F ;convert rjmp target
	ldi	temp2, HIGH(-addr0)
	subi	InstrL, LOW(-addr0)
	sbc	temp, temp2
	andi	temp, 0x0F
	ori	temp, 0xC0
	rjmp	bl_RR2
bl_RR1:
	lpm	InstrL, Z+
	lpm	temp, Z+
bl_RR2:
	rcall	putc
	mov	temp, InstrL
	movw	AddrH:AddrL, Addr2H:Addr2L ;restore address
	adiw	AddrH:AddrL, 2
	rjmp	ret_put
bl_RR:
	
	cpi	temp, 'e'		;'E' chip erase
	brne	bl_e
	ldi	AddrL, LOW((bl_start-PAGESIZE)*2) ;address of last user page
	ldi	AddrH, HIGH((bl_start-PAGESIZE)*2)
bl_e1:
	ldi	temp, 0x03 ;erase page
	rcall	do_spm
	sbiw	AddrH:AddrL, PAGESIZE*2
	brne	bl_e1
	;write "nop" to addr0 and restore "rjmp bl_start" at 0x000
	ldi	InstrL, 0x00
	ldi	InstrH, 0x00
	rcall	firstinstr
	ldi	temp, 0x01 ;load to buffer
	rcall	do_spm
	rcall	writepage
	rjmp	ret_ok
bl_e:

	cpi	temp, 'c'		;'c' write program memory (buffer), low byte
	brne	bl_c
	rcall	getc
	mov	DataL, temp
	rjmp	ret_ok
bl_c:

	cpi	temp, 'C'		;'C' write program memory (buffer), high byte
	brne	bl_CC
	rcall	getc
	mov	DataH, temp
	;*** if Addr = 0: replace the nop at addr0 with the rjmp instr for 0x000 ***
	;***              and (re)write "rjmp bl_start" to 0x000                 ***
	cpi	AddrL, 0x00 ;Addr=0 ?
	cpc	AddrH, AddrL
	brne	bl_CC1
	;read in page, exchange the nop at addr0 with the converted user program rjmp
	movw	InstrH:InstrL, DataH:DataL ;backup instruction
	andi	InstrH, 0x0F ;convert rjmp target
	ldi	temp, HIGH(addr0)
	subi	InstrL, LOW(addr0)
	sbc	InstrH, temp
	andi	InstrH, 0x0F
	ori	InstrH, 0xC0
	rcall	firstinstr ;write to flash
bl_CC1:
	ldi	temp, 0x01 ;store Data (word) in buffer
	rcall	do_spm
	adiw	AddrH:AddrL, 2 ;increment address
	rjmp	ret_ok
bl_CC:

	cpi	temp, 'm'		;'m' issue page write
	brne	bl_m
	ldi	temp, HIGH(bl_start*2) ;do not accept adresses where the bootloader resides
	cpi	AddrL, LOW(bl_start*2)
	cpc	AddrH, temp
	brsh	ret_err
	rcall	writepage
	rjmp	ret_ok
bl_m:
	
;****************************** programmer ******************************

	cpi	temp, 't'		;'t' read supported device codes
	brne	bl_t
	ldi	temp, DevCode
	rcall	putc
	ldi	temp, 0x00
	rjmp	ret_put
bl_t:

	cpi	temp, 'S'		;'S' return software identifier
	brne	bl_SS
	ldi	temp, 'U' ;return "UTSRQPO"
bl_SS1:
	rcall	putc
	subi	temp, 1
	brhc	bl_SS1
	rjmp	ret_put
bl_SS:

	cpi	temp, 'V'		;'V' return software version
	brne	bl_VV
	ldi	temp, 0x30 | VerMaj
	rcall	putc
	ldi	temp, 0x30 | VerMin
	rjmp	ret_put
bl_VV:

	cpi	temp, 'v'		;'v' return hardware version
	brne	bl_v
	ldi	temp, 0x30
	rcall	putc
	rjmp	ret_put
bl_v:

	cpi	temp, 'p'		;'p' return programmer type
	brne	bl_p
	ldi	temp, 'S' ;S = serial
	rjmp	ret_put
bl_p:

	cpi	temp, 'T'		;'T' select device type
	brne	bl_TT
	rcall	getc
	cpi	temp, DevCode
	brne	ret_err
	rjmp	ret_ok
bl_TT:

	cpi	temp, 'P'		;'P' enter programming mode (disable countdown)
	brne	bl_PP
	ldi	count, 0
	rjmp	ret_ok
bl_PP:

	cpi	temp, 'L'		;'L' leave programming mode (re-enable countdown)
	brne	bl_LL
	ldi	count, 1
	rjmp	ret_ok
bl_LL:
	
;****************************** end of commands ******************************
	
ret_err:
	;unknown command or error, return '?'
	ldi	temp, '?'
	rjmp	ret_put

ret_ok:
	;everything went right, return CR
	ldi	temp, 13
ret_put:
	;return the char in temp
	rcall	putc
	rjmp	bl_loop
	
getc:
	;receive a character from RS232
	rcall	countdown
	sbis	USR, RXC
	rjmp	getc
	in	temp, UDR
	rcall	countdown_reset
	ret

putc:
	;send a character via RS232
	rcall	countdown
	sbis	USR, UDRE
	rjmp	putc
	out	UDR, temp
	rcall	countdown_reset
	ret
	
;--------------------
	
countdown:
	;count down and jump to the user program when reached 0
	cpi	count, 0 ;countdown disabled
	breq	return
	sbiw	Count3:Count2, 1
	brne	return
	dec	Count1
	brne	return
	
	;disable UART
	ldi	temp, 0x00
	out	UCSRA, temp
	out	UCSRB, temp
	out	UBRRL, temp
	
	;disable RS232 pullup
	out	PORTD, temp
	
addr0:
	;*****************************************************************
	    ;JUMP TO USER PROGRAM (BOOTLOADER EXITS HERE)
	nop ;gets replaced by the rjmp instr. at 0x000 in the user program
	    ;THIS COMMAND MUST NOT BE IN THE SAME PAGE AS THE SPM COMMAND!
	;*****************************************************************
	
	rjmp	addr0
	
countdown_reset:
	;reset countdown
	ldi	Count3, 0
	ldi	Count2, 0
	ldi	Count1, Count1_init
return:
	;return to the address on top of the stack
	ret
	
;--------------------
	
firstinstr:
	;write instruction in InstrH:InstrL to addr0
	ldi	temp, 0x11 ;clear page buffer
	rcall	do_spm
	ldi	AddrL, LOW(addr0*2) & ~(PAGESIZE*2-1) ;start of the page with "addr0" in it
	ldi	AddrH, HIGH(addr0*2)
	ldi	temp2, PAGESIZE
firstinstr_loop:
	lpm	DataL, Z+
	lpm	DataH, Z+
	sbiw	AddrH:AddrL, 2
	cpi	AddrL, LOW(addr0*2)
	brne	firstinstr_write
	movw	DataH:DataL, InstrH:InstrL
firstinstr_write:
	ldi	temp, 0x01 ;store in buffer
	rcall	do_spm
	adiw	AddrH:AddrL, 2
	dec	temp2
	brne	firstinstr_loop
	;write page
	;ldi	AddrL, LOW(addr0*2) & ~(PAGESIZE*2-1) ;start of the page with "addr0" in it
	;ldi	AddrH, HIGH(addr0*2)
	sbiw	AddrH:AddrL, PAGESIZE*2 ;use starting address again
	rcall	writepage
	ldi	AddrL, 0x00 ;restore address
	ldi	AddrH, 0x00
	ldi	temp2, HIGH(bl_start-1) | 0xC0 ;new command for 0x000: "rjmp bl_start"
	ldi	temp, LOW(bl_start-1)
	movw	DataH:DataL, temp2:temp
	ret
	
;--------------------
	
writepage:
	;erases the page and writes the buffer to it
	andi	AddrL, ~(PAGESIZE*2-1)
	ldi	temp, 0x03 ;erase page
	rcall	do_spm
	ldi	temp, 0x05 ;write page
do_spm:
	;execute the spm action specified in temp
	out	SPMCSR, temp
	spm
	ret
